#ifndef LIB_C
#define LIB_C
/*
 * A bit of a hack, callers will include a .c file for extra
 * performance.  Malloc does the same, for the same reason.
 */

#ifdef USE_DLMALLOC
#define malloc		dlmalloc
#define calloc		dlcalloc
#define free		dlfree
void *dlmalloc(size_t size);
void dlfree(void *ptr);
void *dlcalloc(size_t nmemb, size_t size);
#endif

#if (defined(__i386__) || defined(__amd64__) || defined(__x86_64__))
static inline int fls(int x)
{
	int r;

	asm("bsrl %1,%0\n\t"
	"jnz 1f\n\t"
	"movl $-1,%0\n"
	"1:" : "=r" (r) : "rm" (x));
	return r + 1;
}
#else
#error Must define fls()
#endif

struct mrand {
	uint64_t a[5];
	uint64_t acc;
};

static void mrand_init(struct mrand *r, int seed)
{
	r->a[0] = 0x736f6d6570736575ULL;
	r->a[1] = 0x646f72616e646f6dULL;
	r->a[2] = 0x6c7967656e657261ULL;
	r->a[3] = 0x7465646279746573ULL;
	r->acc = seed;
}


#define ROTL(x,b) (uint64_t)( ((x) << (b)) | ( (x) >> (64 - (b))) )

static uint32_t mrand(struct mrand *r)
{
	uint64_t acc = r->acc + 1;

	//printf("%lx %lx %lx %lx %lx %lx\n", r->a[0], r->a[1], r->a[2], r->a[3], r->a[4], acc);
#define v0 r->a[0]
#define v1 r->a[1]
#define v2 r->a[2]
#define v3 r->a[3]
	v3 ^= acc;
	v0 += v1; v1=ROTL(v1,13); v1 ^= v0; v0=ROTL(v0,32);
	v2 += v3; v3=ROTL(v3,16); v3 ^= v2;
	v0 += v3; v3=ROTL(v3,21); v3 ^= v0;
	v2 += v1; v1=ROTL(v1,17); v1 ^= v2; v2=ROTL(v2,32);
	acc ^= v0;
#undef v0
#undef v1
#undef v2
#undef v3
	r->acc = acc;
	return acc & 0xffffffff;
}

static unsigned mrand_max(struct mrand *r, unsigned max)
{
	unsigned ret;
	unsigned mask = (1 << (fls(max) + 1)) - 1;

	assert(mask >= max);
	do {
		ret = mrand(r) & mask;
	} while (ret > max);
	return ret;
}

#define min(a, b) ((a) < (b) ? (a) : (b))
static size_t alloc_size(struct mrand *rdata)
{
	int size;
	int mask;
	int bits;

	do {
		size = mrand(rdata);
		bits = (size >> 28) & 15;
		bits = min(bits, (size >> 24) & 15);
		bits = min(bits, (size >> 20) & 15);
		mask = (1 << bits) - 1;
		size &= mask;
		size *= sizeof(void *);
		size += 16; /* minimum size */
	} while (size > max_size);
	//printf("%8x\n", size);
	return size;
}

static void *check_alloc(size_t size, void *fill)
{
	size_t len = size / sizeof(void *);
	void **buf;
	int i;

	buf = malloc(size);
	buf[0] = (void *)size;
	for (i = 1; i < len; i++) {
		buf[i] = fill;
	}
	return buf;
}

static size_t check_mem(void *_buf, void *fill)
{
	void **buf = _buf;
	size_t size, len;
	int i;

	assert(buf);
	size = (size_t)(buf[0]);
	len = size / sizeof(void *);
	for (i = 1; i < len; i++) {
		assert(buf[i] == fill);
	}

	return size;
}

static inline size_t check_free(void *buf, void *fill)
{
	size_t size = check_mem(buf, fill);

	free(buf);
	return size;
}
#endif
