/*
 * Moves memory between threads to maximize lock contention
 */
#include <assert.h>
#include <pthread.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/* These should become parameters */
#define default_iterations	(1024 * 1024 * 1024)
#define default_max_size	(64 * 1024)
#define default_max_mem		(1024 * 1024 * 1024)

#define iterations	(1024 * 1024)
#define max_size	(default_max_size)
#define max_mem		(default_max_mem)

#define thread_mem	(max_mem / no_threads)
#define array_size	(thread_mem / 1024)

#include "lib.c"

#define max_threads	(1024)
#define no_lists	(16 * max_threads)
static int no_threads;
void *cross_lists[no_lists];

static size_t cross_free(struct mrand *r, void *buf, void *fill)
{
	size_t size = check_mem(buf, fill);
	unsigned int index = mrand_max(r, no_lists - 1);
	void *oldbuf;
	int i = 0;

	assert(index < no_lists);
	do {
		oldbuf = cross_lists[index];
		assert(i++ < 1000);
	} while (!__sync_bool_compare_and_swap(&cross_lists[index], oldbuf, buf));
	free(oldbuf);
	return size;
}

static void *thread_func(void *_seed)
{
	int seed = (long)_seed;
	struct mrand rdata;
	int i;
	void **array, **slot;
	size_t size, total_mem = array_size * sizeof(void*);

	mrand_init(&rdata, seed);
	assert(array = calloc(array_size, sizeof(void*)));
	for (i = iterations; i; i--) {
		slot = &array[mrand_max(&rdata, array_size - 1)];
		if (*slot) {
			total_mem -= cross_free(&rdata, *slot, slot);
			*slot = NULL;
		} else if (total_mem < thread_mem) {
			size = alloc_size(&rdata);
			total_mem += size;
			*slot = check_alloc(size, slot);
		}
	}
	for (i = 0; i < array_size; i++) {
		slot = array[i];
		if (slot) {
			free(slot);
		}
	}
	free(array);
	return NULL;
}

static void main_thread(void)
{
	pthread_t thread[no_threads];
	int i;

	for (i = 0; i < no_threads; i++) {
		pthread_create(&thread[i], NULL, thread_func, (void *)(long)i);
	}
	for (i = 0; i < no_threads; i++) {
		pthread_join(thread[i], NULL);
	}
}

int main(void)
{
	long cpus = sysconf(_SC_NPROCESSORS_ONLN);
	assert(cpus > 0);
	no_threads = cpus * 2;
	assert(no_threads <= max_threads);
	main_thread();
	return 0;
}
